/*
* Copyright (C) 2014 Andrew Reid and the ModelGUI Project <http://mgui.wikidot.com>
* 
* This file is part of ModelGUI[core] (mgui-core).
* 
* ModelGUI[core] is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* ModelGUI[core] is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with ModelGUI[core]. If not, see <http://www.gnu.org/licenses/>.
*/

package mgui.geometry;

import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;

import javax.vecmath.Point2f;
import javax.vecmath.Tuple2f;

import mgui.geometry.util.GeometryFunctions;
import mgui.interfaces.gui.InterfaceProgressBar;
import mgui.interfaces.xml.XMLFunctions;

import org.xml.sax.Attributes;

/********************************************
 * Represents a series of connected 2D points, defining either a polyline
 * or a closed polygon.
 * 
 * @author Andrew Reid
 * @since 1.0
 * @version 1.0
 *
 */
public class Polygon2D extends Shape2D {
	//variables for: nodes, as ArrayList

	public ArrayList<Point2f> nodes;
	
	public Polygon2D(){
		nodes = new ArrayList<Point2f>();
	}
	
	public Polygon2D(ArrayList<Point2f> nodeList){
		nodes = nodeList;
	}
	
	@Override
	public Point2f getCenterPt(){
		return Polygon2D.getCenterOfGravity(this);
	}
	
	@Override
	public Point2f getVertex(int i){
		return nodes.get(i);
	}
	
	public Point2f getNextNode(int i){
		if (i == nodes.size() - 1)
			return nodes.get(0);
		return nodes.get(i + 1);
	}
	
	@Override
	public int getSize(){
		return nodes.size();
	}
	
	public Point2f getPrevNode(int i){
		if (i == 0)
			return nodes.get(nodes.size() - 1);
		return nodes.get(i - 1);
	}
	
	public Point2f getFirstNode(){
		return nodes.get(0);
	}
	
	public Point2f getLastNode(){
		return nodes.get(nodes.size() - 1);
	}
	
	public void addNode(Point2f p){
		nodes.add(p);
	}
	
	@Override
	public boolean contains(Point2f pt){
		return GeometryFunctions.isInternalConcave(this, pt);
	}
	
	/**
	 * @param thisPoly
	 * @return a Point2f object that is the center of gravity for this polygon
	 */
	public static Point2f getCenterOfGravity(Polygon2D thisPoly){
		//sum x's and y's
		float sumX = 0, sumY = 0;
		int polySize = thisPoly.nodes.size();
		for (int i = 0; i < polySize; i++){
			sumX += thisPoly.nodes.get(i).x;
			sumY += thisPoly.nodes.get(i).y;
		}
		//divide by no. nodes
		return new Point2f(sumX / polySize, sumY / polySize);
	}
	
	@Override
	public Object clone(){
		ArrayList<Point2f> newNodes = new ArrayList<Point2f>(nodes.size());
		for (int i = 0; i < nodes.size(); i++)
			newNodes.add(new Point2f((Tuple2f)nodes.get(i).clone()));
		if (newNodes.get(0).x == Float.NaN)
			return null;
		return new Polygon2D(newNodes);
	}
	
	@Override
	public Point2f getProximityPoint(Point2f thisPt){
		int j;
		float nearest = Float.MAX_VALUE;
		float thisDist;
		LineSegment2D thisLine;
		Point2f retPoint = null;
		
		for (int i = 0; i < nodes.size(); i++){
		j = i + 1;
		if (j == nodes.size()) j = 0;
		thisLine = new LineSegment2D(nodes.get(i), nodes.get(j));
			thisDist = thisLine.getProximity(thisPt);
			if (nearest > thisDist && thisDist > 0){
				nearest = thisDist;
				retPoint = thisLine.getProximityPoint(thisPt);
				}
			}
		return retPoint;
	}
	
	@Override
	public ArrayList<Point2f> getVertices(){
		return nodes;
	}
	
	public void setVertices(ArrayList<Point2f> vertices){
		nodes = new ArrayList<Point2f>(vertices.size());
		for (int i = 0; i < vertices.size(); i++)
			nodes.add(new Point2f(vertices.get(i)));
	}
	
	@Override
	public float getProximity(Point2f thisPt){
		//need to determine perpendicular distance from each edge to this point
		int j;
		float nearest = Float.MAX_VALUE;
		float thisDist;
		
		for (int i = 0; i < nodes.size(); i++){
		j = i + 1;
		if (j == nodes.size()) j = 0;
			thisDist = new LineSegment2D(nodes.get(i), nodes.get(j)).getProximity(thisPt);
			if (nearest > thisDist && thisDist > 0)
				nearest = thisDist;
			}
		return nearest;
	}
	
	public void writeXML(int tab, Writer writer, InterfaceProgressBar progress_bar) throws IOException{
		
		String _tab = XMLFunctions.getTab(tab);
		String _tab2 = XMLFunctions.getTab(tab + 1);
		writer.write(_tab + "<Nodes n='" + nodes.size() + "'>\n");
		
		for (int i = 0; i < nodes.size(); i++)
			writer.write(_tab2 + XMLFunctions.getXMLForPoint2f(nodes.get(i), 8) + "\n");
		
		writer.write(_tab + "</Nodes>\n");
		
	}
	
	boolean start_xml = false;
	
	@Override
	public void handleXMLElementEnd(String localName) {
		if (localName.equals(getLocalName())){
			start_xml = false;
			}
	}

	@Override
	public void handleXMLElementStart(String localName, Attributes attributes, XMLType type) {
		
		if (localName.equals(getLocalName())){
			start_xml = true;
			}
		
	}

	@Override
	public void handleXMLString(String s) {
		if (!start_xml) return;
		
		addNode(XMLFunctions.getPoint2f(s));
		
	}
	
}