/*
* Copyright (C) 2014 Andrew Reid and the ModelGUI Project <http://mgui.wikidot.com>
* 
* This file is part of ModelGUI[core] (mgui-core).
* 
* ModelGUI[core] is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* ModelGUI[core] is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with ModelGUI[core]. If not, see <http://www.gnu.org/licenses/>.
*/

package mgui.io.domestic.shapes;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.StringTokenizer;

import mgui.geometry.Polygon3D;
import mgui.interfaces.InterfaceSession;
import mgui.interfaces.ProgressUpdater;
import mgui.interfaces.shapes.PolygonSet3DInt;
import mgui.io.FileLoader;
import mgui.io.InterfaceIOOptions;
import foxtrot.Job;
import foxtrot.Worker;


public class PolygonSet3DLoader extends FileLoader {

	public boolean skip_lines = false;
	public int skip = 1;
	public boolean skip_min_nodes = false;
	public int min_nodes = 1;
	
	@Override
	public boolean load(InterfaceIOOptions options, ProgressUpdater progress_bar){
		PolygonSet3DInputOptions _options = (PolygonSet3DInputOptions)options;
		boolean success = true;
		
		for (int i = 0; i < _options.getFiles().length; i++){
			setFile(_options.getFiles()[i]);
			try{
				skip_lines = _options.skip_lines;
				skip = _options.skip;
				skip_min_nodes = _options.skip_min_nodes;
				min_nodes = _options.min_nodes;
				PolygonSet3DInt set = loadPolygonSet(progress_bar);
				set.setName(_options.names[i]);
				if (set == null)
					success = false;
				else{
					System.out.print("Adding Polygon3D set..");
					_options.shapeSet.addShape(set);
					InterfaceSession.log("done.");
					}
			}catch (IOException e){
				e.printStackTrace();
				success = false;
				}
			}
		
		return success;
	}
	
	@Override
	public Object loadObject(ProgressUpdater progress_bar, InterfaceIOOptions options) throws IOException{
		return loadPolygonSet(progress_bar);
	}
	
	public PolygonSet3DInt loadPolygonSet(final ProgressUpdater progress_bar) throws IOException{
		
		if (dataFile == null)
			throw new IOException(getClass().getSimpleName() + ": No input file set.");
		
		PolygonSet3DInt polygon_set = null;
		
		if (progress_bar != null){
			polygon_set = (PolygonSet3DInt)Worker.post(new Job(){
				@Override
				public PolygonSet3DInt run(){
					try{
						return loadPolygonSetBlocking(progress_bar);
					}catch (IOException e){
						e.printStackTrace();
						return null;
						}
				}
			});
		}else{
			polygon_set = loadPolygonSetBlocking(progress_bar);
			}
		
		return polygon_set;
	}
	
	protected PolygonSet3DInt loadPolygonSetBlocking(ProgressUpdater progress_bar) throws IOException{
		
		//load in domestic format
		BufferedReader reader = new BufferedReader(new FileReader(dataFile));
		String line = reader.readLine();
		if (line == null)
			throw new IOException("PolygonSet3DLoader: unexpected end of file in '" + dataFile.getAbsolutePath() + "'.");
		
		int n_polys = Integer.valueOf(line);
		
		if (progress_bar != null){
			progress_bar.setMessage("Loading polygons from '" + dataFile.getName() + "': ");
			progress_bar.setMinimum(0);
			progress_bar.setMaximum(n_polys);
			progress_bar.reset();
			}
		
		PolygonSet3DInt polygons = new PolygonSet3DInt();
		
		for (int i = 0; i < n_polys; i++){
			line = reader.readLine();
			while (line != null && line.length() == 0)
				line = reader.readLine();
			if (line == null)
				throw new IOException("PolygonSet3DLoader: unexpected end of file in '" + dataFile.getAbsolutePath() + "'.");
			int n = Integer.valueOf(line);
			
			Polygon3D polygon =  new Polygon3D();
			float[] nodes = new float[n * 3];
			for (int j = 0; j < n; j++){
				line = reader.readLine();
				if (line == null)
					throw new IOException("PolygonSet3DLoader: unexpected end of file in '" + dataFile.getAbsolutePath() + "'.");

				StringTokenizer tokens = new StringTokenizer(line);
				nodes[j * 3] = Float.valueOf(tokens.nextToken());
				nodes[(j * 3) + 1] = Float.valueOf(tokens.nextToken());
				nodes[(j * 3) + 2] = Float.valueOf(tokens.nextToken());
				}
			
			polygon.nodes = nodes;
			polygon.n = n;
			
			if (progress_bar != null)
				progress_bar.update(i);
			polygons.addShape(polygon);
			}

		reader.close();
		return polygons;
	}
	
}